#
# This file is part of Config-Model
#
# This software is Copyright (c) 2015 by Dominique Dumont.
#
# This is free software, licensed under:
#
#   The GNU Lesser General Public License, Version 2.1, February 1999
#
[
          {
            'class_description' => 'options valid for all types of file systems.',
            'name' => 'MyFstab::CommonOptions',
            'element' => [
                           'async',
                           {
                             'value_type' => 'boolean',
                             'type' => 'leaf'
                           },
                           'atime',
                           {
                             'value_type' => 'boolean',
                             'type' => 'leaf'
                           },
                           'auto',
                           {
                             'value_type' => 'boolean',
                             'type' => 'leaf'
                           },
                           'dev',
                           {
                             'value_type' => 'boolean',
                             'type' => 'leaf'
                           },
                           'exec',
                           {
                             'value_type' => 'boolean',
                             'type' => 'leaf'
                           },
                           'group',
                           {
                             'value_type' => 'boolean',
                             'type' => 'leaf'
                           },
                           'mand',
                           {
                             'value_type' => 'boolean',
                             'type' => 'leaf'
                           },
                           'user',
                           {
                             'value_type' => 'boolean',
                             'help' => {
                                         '1' => 'user can mount the file system',
                                         '0' => 'Only root can mount the file system'
                                       },
                             'type' => 'leaf'
                           },
                           'defaults',
                           {
                             'value_type' => 'boolean',
                             'help' => {
                                         '1' => 'option equivalent to rw, suid, dev, exec, auto, nouser, and async'
                                       },
                             'type' => 'leaf'
                           },
                           'rw',
                           {
                             'value_type' => 'boolean',
                             'help' => {
                                         '0' => 'read-only file system'
                                       },
                             'type' => 'leaf'
                           }
                         ]
          }
        ]
;
